import torch
from torch import tensor, Tensor
import numpy as np
from torch import tensor
import torch.nn as nn
import torch
import math

from typing import Union, Tuple, Optional, List
from .._unet_utils import is_contiguous as is_cont


def A(*a):
    return np.array(a[0]) if len(a) == 1 else [np.array(o) for o in a]


imagenet_stats = A([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

mean = 255 * np.array(imagenet_stats[0], dtype=np.float32)
std = 255 * np.array(imagenet_stats[1], dtype=np.float32)

norm = lambda x: (x - mean) / std
denorm = lambda x: x * std + mean


def scale_batch(
    image_batch, model_info, normalization_stats=None, break_extract_bands=False
):
    if normalization_stats is None:
        normalization_stats = model_info.get("NormalizationStats", None)
    if break_extract_bands:
        # Only for change detection
        # if subset of extract bands are specified fix this.
        n_bands = len(model_info["ExtractBands"]) // 2
        band_min_values = np.array(normalization_stats["band_min_values"])[
            model_info["ExtractBands"][:n_bands]
        ].reshape(1, -1, 1, 1)
        band_max_values = np.array(normalization_stats["band_max_values"])[
            model_info["ExtractBands"][:n_bands]
        ].reshape(1, -1, 1, 1)
    else:
        if normalization_stats is None:
            modtype = model_info.get("InferenceFunction", None)
            if modtype == "ArcGISImageTranslation.py":
                band_min_values = np.full((image_batch.shape[1],), 0).reshape(
                    1, -1, 1, 1
                )
                band_max_values = np.full((image_batch.shape[1],), 255).reshape(
                    1, -1, 1, 1
                )
        else:
            band_min_values = np.array(normalization_stats["band_min_values"])[
                model_info["ExtractBands"]
            ].reshape(1, -1, 1, 1)
            band_max_values = np.array(normalization_stats["band_max_values"])[
                model_info["ExtractBands"]
            ].reshape(1, -1, 1, 1)
    img_scaled = (image_batch - band_min_values) / (band_max_values - band_min_values)
    return img_scaled


def normalize_batch(
    image_batch, model_info=None, normalization_stats=None, prithvi=False
):
    if normalization_stats is None:
        normalization_stats = model_info.get("NormalizationStats", None)
    scaled_mean_values = np.array(normalization_stats["scaled_mean_values"])[
        model_info["ExtractBands"]
    ].reshape(1, -1, 1, 1)
    scaled_std_values = np.array(normalization_stats["scaled_std_values"])[
        model_info["ExtractBands"]
    ].reshape(1, -1, 1, 1)
    if prithvi:
        img_normed = (image_batch - scaled_mean_values) / scaled_std_values
        return img_normed
    else:
        img_scaled = scale_batch(image_batch, model_info)
        img_normed = (img_scaled - scaled_mean_values) / scaled_std_values
        return img_normed


def ts_normalization(x, m, s):
    x = np.rollaxis(x, 2)  # TxCxS -> SxTxC
    x = (x - m) / s
    return torch.tensor(x[None, :, :, :, None])


def rescale_batch(
    image_batch, model_info, normalization_stats=None, break_extract_bands=False
):
    if normalization_stats is None:
        modtype = model_info.get("InferenceFunction", None)
        if modtype == "ArcGISImageTranslation.py":
            min_values = np.full((image_batch.shape[1],), 0).reshape(1, -1, 1, 1)
            max_values = np.full((image_batch.shape[1],), 255).reshape(1, -1, 1, 1)
    else:
        min_values = np.array(normalization_stats["band_min_values"])
        max_values = np.array(normalization_stats["band_max_values"])

        x = image_batch
        if x.shape[1] > min_values.shape[0]:
            res = x.shape[1] - min_values.shape[0]
            last_val = torch.tensor([min_values[min_values.shape[0] - 1]])
            for i in range(res):
                min_values = np.concatenate((min_values, last_val), axis=0)
        if x.shape[1] > max_values.shape[0]:
            res = x.shape[1] - max_values.shape[0]
            last_val = torch.tensor([max_values[max_values.shape[0] - 1]])
            for i in range(res):
                max_values = np.concatenate((max_values, last_val), axis=0)

        min_values = min_values.reshape(1, -1, 1, 1)
        max_values = max_values.reshape(1, -1, 1, 1)
    img_rescaled = (((image_batch + 1) / 2) * (max_values - min_values)) + min_values
    return img_rescaled


def pred2dict(bb_np, score, cat_str, c):
    # convert to top left x,y bottom right x,y
    return {
        "x1": bb_np[1],
        "x2": bb_np[3],
        "y1": bb_np[0],
        "y2": bb_np[2],
        "score": score,
        "category": cat_str,
        "class": c,
    }


def to_np(x):
    return x.cpu().numpy()


def load_weights(m, p):
    if p.split(".")[-1] == "h5":
        sd = torch.load(p, map_location=lambda storage, loc: storage)
    elif p.split(".")[-1] == "pth":
        # sd = torch.load(p)['model']
        sd = torch.load(p, map_location=lambda storage, loc: storage)["model"]
    m.load_state_dict(sd)
    return m


def predict_(model, images, device):
    model = model.to(device)
    images = tensor(images).to(device).float()
    clas, bbox = model(images)
    return clas, bbox


def hw2corners(ctr, hw):
    return torch.cat([ctr - hw / 2, ctr + hw / 2], dim=1)


def actn_to_bb(actn, anchors, grid_sizes):
    actn_bbs = torch.tanh(actn)
    actn_centers = (actn_bbs[:, :2] / 2 * grid_sizes) + anchors[:, :2]
    actn_hw = (actn_bbs[:, 2:] / 2 + 1) * anchors[:, 2:]
    return hw2corners(actn_centers, actn_hw)


def nms(boxes, scores, overlap=0.5, top_k=100):
    keep = scores.new(scores.size(0)).zero_().long()
    if boxes.numel() == 0:
        return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    area = torch.mul(x2 - x1, y2 - y1)
    v, idx = scores.sort(0)  # sort in ascending order
    idx = idx[-top_k:]  # indices of the top-k largest vals
    xx1 = boxes.new()
    yy1 = boxes.new()
    xx2 = boxes.new()
    yy2 = boxes.new()
    w = boxes.new()
    h = boxes.new()

    count = 0
    while idx.numel() > 0:
        i = idx[-1]  # index of current largest val
        keep[count] = i
        count += 1
        if idx.size(0) == 1:
            break
        idx = idx[:-1]  # remove kept element from view
        # load bboxes of next highest vals
        torch.index_select(x1, 0, idx, out=xx1)
        torch.index_select(y1, 0, idx, out=yy1)
        torch.index_select(x2, 0, idx, out=xx2)
        torch.index_select(y2, 0, idx, out=yy2)
        # store element-wise max with next highest score
        xx1 = torch.clamp(xx1, min=x1[i])
        yy1 = torch.clamp(yy1, min=y1[i])
        xx2 = torch.clamp(xx2, max=x2[i])
        yy2 = torch.clamp(yy2, max=y2[i])
        w.resize_as_(xx2)
        h.resize_as_(yy2)
        w = xx2 - xx1
        h = yy2 - yy1
        # check sizes of xx1 and xx2.. after each iteration
        w = torch.clamp(w, min=0.0)
        h = torch.clamp(h, min=0.0)
        inter = w * h
        # IoU = i / (area(a) + area(b) - i)
        rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
        union = (rem_areas - inter) + area[i]
        IoU = inter / union  # store result in iou
        # keep only elements with an IoU <= overlap
        idx = idx[IoU.le(overlap)]
    return keep, count


try:

    @torch.jit.script
    def nms_jit(
        boxes, scores, overlap: float = 0.5, top_k: int = 100
    ) -> Tuple[Tensor, Tensor]:
        keep = torch.zeros(scores.size(0), dtype=torch.long).to(scores.device)

        count = 0
        if boxes.numel() == 0:
            return keep, torch.tensor([count]).to(scores.device).int()
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]
        area = torch.mul(x2 - x1, y2 - y1)
        v, idx = scores.sort(0)  # sort in ascending order
        limiter = 0 if top_k == 0 else len(idx) - top_k  # TODO:here
        idx = idx[limiter:]  # indices of the top-k largest vals
        xx1 = torch.empty((0,), dtype=boxes.dtype).to(scores.device)
        yy1 = torch.empty((0,), dtype=boxes.dtype).to(scores.device)
        xx2 = torch.empty((0,), dtype=boxes.dtype).to(scores.device)
        yy2 = torch.empty((0,), dtype=boxes.dtype).to(scores.device)
        w = torch.empty((0,), dtype=boxes.dtype).to(scores.device)
        h = torch.empty((0,), dtype=boxes.dtype).to(scores.device)

        while idx.numel() > 0:
            i = idx[len(idx) - 1]  # index of current largest val
            keep[count] = i
            count += 1
            if idx.size(0) == 1:
                break
            idx = (
                idx[: len(idx) - 1].clone().to(scores.device)
            )  # remove kept element from view
            # load bboxes of next highest vals
            torch.index_select(x1, 0, idx, out=xx1)
            torch.index_select(y1, 0, idx, out=yy1)
            torch.index_select(x2, 0, idx, out=xx2)
            torch.index_select(y2, 0, idx, out=yy2)
            # store element-wise max with next highest score
            xx1 = torch.clamp(xx1, min=x1[i])
            yy1 = torch.clamp(yy1, min=y1[i])
            xx2 = torch.clamp(xx2, max=x2[i])
            yy2 = torch.clamp(yy2, max=y2[i])
            w.resize_as_(xx2)
            h.resize_as_(yy2)
            w = xx2 - xx1
            h = yy2 - yy1
            # check sizes of xx1 and xx2.. after each iteration
            w = torch.clamp(w, min=0.0)
            h = torch.clamp(h, min=0.0)
            inter = w * h
            # IoU = i / (area(a) + area(b) - i)
            rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
            union = (rem_areas - inter) + area[i]
            IoU = inter / union  # store result in iou
            # keep only elements with an IoU <= overlap
            idx = idx[IoU.le(overlap)]

        return keep, torch.tensor([count]).to(scores.device).int()

except:
    print("\ntorch not available\n")


def get_nms_preds(b_clas, b_bb, idx, anchors, grid_sizes, classes, nms_overlap, thres):
    a_ic = actn_to_bb(b_bb[idx], anchors, grid_sizes)
    clas_pr, clas_ids = b_clas[idx].max(1)
    clas_pr = clas_pr.sigmoid()

    conf_scores = b_clas[idx].sigmoid().t().data

    out1, out2, cc = [], [], []
    for cl in range(1, len(conf_scores)):
        c_mask = conf_scores[cl] > thres
        if c_mask.sum() == 0:
            continue
        scores = conf_scores[cl][c_mask]
        l_mask = c_mask.unsqueeze(1).expand_as(a_ic)
        boxes = a_ic[l_mask].view(-1, 4)
        ids, count = nms(
            boxes.data, scores, nms_overlap, 50
        )  # FIX- NMS overlap hardcoded
        ids = ids[:count]
        out1.append(scores[ids])
        out2.append(boxes.data[ids])
        cc.append([cl] * count)
    if cc == []:
        cc = [[0]]
    cc = tensor(np.concatenate(cc))
    if out1 == []:
        out1 = [torch.Tensor()]
    out1 = torch.cat(out1)
    if out2 == []:
        out2 = [torch.Tensor()]
    out2 = torch.cat(out2)
    bbox, clas, prs, thresh = out2, cc, out1, thres  # FIX- hardcoded threshold
    return predictions(
        bbox, to_np(clas), to_np(prs) if prs is not None else None, thresh, classes
    )


def predictions(
    bbox, clas=None, prs=None, thresh=0.3, classes=None
):  # FIX- take threshold from user
    # bb = [bb_hw(o) for o in bbox.reshape(-1,4)]
    if len(classes) is None:
        raise Exception("Classes are None")
    else:
        classes = ["bg"] + classes
    bb = bbox
    if prs is None:
        prs = [None] * len(bb)
    if clas is None:
        clas = [None] * len(bb)
    predictions = []
    for i, (b, c, pr) in enumerate(zip(bb, clas, prs)):
        if (b[2] > 0) and (pr is None or pr > thresh):
            cat_str = classes[c]
            score = pr * 100
            bb_np = to_np(b).astype("float64")
            predictions.append(pred2dict(bb_np, score, cat_str, c))
    return predictions


def detect_objects_image_space(
    model, tiles, anchors, grid_sizes, device, classes, nms_overlap, thres, model_info
):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]
    if "NormalizationStats" in model_info:
        img_normed = normalize_batch(tiles, model_info)
    else:
        img_normed = norm(tiles.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)

    clas, bbox = predict_(model, img_normed, device)

    preds = {}

    for batch_idx in range(bbox.size()[0]):
        preds[batch_idx] = get_nms_preds(
            clas, bbox, batch_idx, anchors, grid_sizes, classes, nms_overlap, thres
        )

    batch_size = bbox.size()[0]
    side = math.sqrt(batch_size)

    num_boxes = 0
    for batch_idx in range(batch_size):
        num_boxes = num_boxes + len(preds[batch_idx])

    bounding_boxes = np.empty(shape=(num_boxes, 4), dtype=float)
    scores = np.empty(shape=(num_boxes), dtype=float)
    classes = np.empty(shape=(num_boxes), dtype=np.uint8)

    idx = 0
    for batch_idx in range(batch_size):
        i, j = batch_idx // side, batch_idx % side

        for pred in preds[batch_idx]:
            bounding_boxes[idx, 0] = (pred["y1"] + i) * tile_height
            bounding_boxes[idx, 1] = (pred["x1"] + j) * tile_width
            bounding_boxes[idx, 2] = (pred["y2"] + i) * tile_height
            bounding_boxes[idx, 3] = (pred["x2"] + j) * tile_width
            scores[idx] = pred["score"]
            classes[idx] = pred["class"]
            idx = idx + 1

    return bounding_boxes, scores, classes


def segment_image(model, images, device, predict_bg, model_info):
    model = model.to(device)
    normed_batch_tensor = tensor(images).to(device).float()
    output = model(normed_batch_tensor)
    ignore_mapped_class = model_info.get("ignore_mapped_class", [])
    for k in ignore_mapped_class:
        output[:, k] = output.min() - 1
    if predict_bg:
        return output.max(dim=1)[1]
    else:
        output[:, 0] = output.min() - 1
        return output.max(dim=1)[1]


def superres_image(model, images, device):
    model = model.to(device)
    normed_batch_tensor = tensor(images).to(device).float()
    output = model(normed_batch_tensor)
    return output


def cyclegan_image(model, images, device, direction):
    model = model.to(device)
    normed_batch_tensor = tensor(images).to(device).float()
    if direction == "BtoA":
        output = model.G_A(normed_batch_tensor)
    else:
        output = model.G_B(normed_batch_tensor)
    return output


def pix2pix_image(model, images, device, label_nc=0):
    from .._pix2pix_hd_utils import encode_input

    model = model.to(device)
    normed_batch_tensor = tensor(images).to(device).float()
    if label_nc:
        normed_batch_tensor, _, _, _ = encode_input(normed_batch_tensor, label_nc)
    output = model.G(normed_batch_tensor)
    return output


def remap(tensor, idx2pixel):
    modified_tensor = torch.zeros_like(tensor)
    for id, pixel in idx2pixel.items():
        modified_tensor[tensor == id] = pixel
    return modified_tensor


def wnet_cgan_image(model, images_a, images_b, device):
    model = model.to(device)
    normed_batch_tensor_a, normed_batch_tensor_b = (
        tensor(images_a).to(device).float(),
        tensor(images_b).to(device).float(),
    )
    output = model.G(normed_batch_tensor_a, normed_batch_tensor_b)
    return output


def pixel_classify_image(model, tiles, device, classes, predict_bg, model_info):
    class_values = [clas["Value"] for clas in model_info["Classes"]]
    is_contiguous = is_cont([0] + class_values)

    if not is_contiguous:
        pixel_mapping = [0] + class_values
        idx2pixel = {i: d for i, d in enumerate(pixel_mapping)}

    tile_height, tile_width = tiles.shape[2], tiles.shape[3]
    if "NormalizationStats" in model_info:
        img_normed = normalize_batch(tiles, model_info)
    else:
        img_normed = norm(tiles.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)
    semantic_predictions = segment_image(
        model, img_normed, device, predict_bg, model_info
    )
    if not is_contiguous:
        semantic_predictions = remap(semantic_predictions, idx2pixel)
    return semantic_predictions


def pixel_classify_superres_image(
    model, tiles, device, model_info, sampling, n_timestep
):
    import numpy as np

    tiles, is_multispec = tensor(tiles), model_info.get("is_multispec")
    modarch = model_info.get("ModelArch", "UNet")
    if modarch == "SR3":
        norm_stats_a = model_info.get("image_stats", None)
        norm_stats_b = model_info.get("image_stats2", None)
        model_info["ExtractBands"] = list(range(tiles.shape[1]))

        if model_info.get("is_multispec"):
            img_scaled = scale_batch(tiles, model_info, norm_stats_a)
        else:
            img_scaled = tiles / 255
        img_normed = -1 + 2 * img_scaled

        kwargs = model_info.get("Kwargs")

        if sampling == "ddim":
            nstp = {"n_timestep": kwargs.get("n_timestep", 1000)}
        else:
            nstp = {"n_timestep": n_timestep}
        combkwargs = {**kwargs, **nstp}
        model.set_new_noise_schedule(device, **combkwargs)

        model = model.to(device)
        normed_batch_tensor = tensor(img_normed).to(device).float()

        if sampling == "ddim":
            superres_predictions = model.super_resolution(
                x_in=normed_batch_tensor,
                continous=False,
                sampling_timesteps=n_timestep,
                ddim_sampling_eta=1,
                sampling="ddim",
            )
        else:
            superres_predictions = (
                model.super_resolution(x_in=normed_batch_tensor, continous=True)
            )[-normed_batch_tensor.shape[0] :]

        min_values = (
            torch.tensor(norm_stats_b["band_min_values"], device=device)
            .unsqueeze(0)
            .unsqueeze(-1)
            .unsqueeze(-1)
        )
        max_values = (
            torch.tensor(norm_stats_b["band_max_values"], device=device)
            .unsqueeze(0)
            .unsqueeze(-1)
            .unsqueeze(-1)
        )

        superres_predictions = (
            superres_predictions.add_(1)
            .div_(2)
            .mul_(max_values - min_values)
            .add_(min_values)
        )
    else:
        if is_multispec:
            n_mean, n_std = np.array(
                model_info.get("image_stats")[0], dtype=np.float32
            ), np.array(model_info.get("image_stats")[1], dtype=np.float32)
            dn_mean, dn_std = (
                model_info.get("image_stats2")[0],
                model_info.get("image_stats2")[1],
            )
            nrm = lambda x, m, s: (x - m) / s
            img_normed = nrm(tiles.permute(0, 2, 3, 1), n_mean, n_std).permute(
                0, 3, 1, 2
            )
        else:
            img_normed = norm(tiles.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
            dn_mean, dn_std = imagenet_stats[0], imagenet_stats[1]
        superres_predictions = superres_image(model, img_normed, device)
        superres_predictions = (
            superres_predictions
            * torch.tensor(dn_std).view(1, -1, 1, 1).to(superres_predictions)
        ) + torch.tensor(dn_mean).view(1, -1, 1, 1).to(superres_predictions)
    return superres_predictions


def pixel_classify_cyclegan_image(model, tiles, device, direction, model_info):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]
    num_channel = model_info.get("n_intput_channel", None)
    if num_channel == None:
        num_channel = model_info.get("n_channel", None)
    num_channel_tar = model_info.get("n_channel", None)
    if direction == "BtoA":
        norm_stats_a = model_info.get("NormalizationStats_b", None)
    else:
        norm_stats_a = model_info.get("NormalizationStats", None)
    img_scaled = scale_batch(tiles, model_info, norm_stats_a)
    img_normed = -1 + 2 * img_scaled
    if img_normed.shape[1] < num_channel:
        cont = []
        for j in range(img_normed.shape[0]):
            tile = img_normed[j, :, :, :]
            last_tile = np.expand_dims(tile[tile.shape[0] - 1, :, :], 0)
            res = abs(num_channel - tile.shape[0])
            for i in range(res):
                tile = np.concatenate((tile, last_tile), axis=0)
            cont.append(tile)
        img_normed = np.stack(cont, axis=0)

    cyclegan_predictions = cyclegan_image(model, img_normed, device, direction)
    if direction == "BtoA":
        norm_stats_b = model_info.get("NormalizationStats", None)
    else:
        norm_stats_b = model_info.get("NormalizationStats_b", None)
    cyclegan_predictions = cyclegan_predictions.detach().cpu().numpy()
    if num_channel_tar == None:
        cyclegan_predictions = torch.tensor(
            rescale_batch(cyclegan_predictions, model_info, norm_stats_b)
        )
    else:
        if direction == "BtoA":
            num_channel_tar_rev = model_info.get("n_channel_rev", None)
            cyclegan_predictions = torch.tensor(
                rescale_batch(cyclegan_predictions, model_info, norm_stats_b)
            )[:, :num_channel_tar_rev, :, :]
        else:
            cyclegan_predictions = torch.tensor(
                rescale_batch(cyclegan_predictions, model_info, norm_stats_b)
            )[:, :num_channel_tar, :, :]
    return cyclegan_predictions


def pixel_classify_ts_image(model, tiles, device, model_info):
    tiles = torch.tensor(tiles)
    tile_height, tile_width, nch = tiles.shape[2], tiles.shape[3], tiles.shape[1]
    means = np.array((model_info.get("mean_norm_stats", None))["mean_stats"])
    stds = np.array((model_info.get("std_norm_stats", None))["std_stats"])
    tile_stack = lambda lst: torch.cat(lst, axis=0)
    img_arr = tile_stack([tile.permute(1, 2, 0)[None, :, :, :] for tile in tiles])
    timeseries_arr = tile_stack(
        [
            torch.reshape(tile_arr, (1, tile_height * tile_width, nch))
            for tile_arr in img_arr
        ]
    )
    ntemp = int(model_info.get("n_temporal", None))
    nchannel = int(model_info.get("n_channel", None))
    ntemp_infer = model_info.get("timestep_infer", None)
    nchannel_infer = model_info.get("channels_infer", None)
    class_dict = model_info.get("Class_mapping", None)
    num_class_dict = model_info.get("Num_class_mapping", None)
    convertmap = model_info.get("convertmap", None)
    bandidx = model_info.get("bandindex", None)
    timeidx = model_info.get("timeindex", None)

    if num_class_dict:
        pixel_num_class_mapping = num_class_dict
    else:
        pixel_num_class_mapping = class_dict

    if bandidx or timeidx:
        final = tile_stack(
            [
                torch.reshape(
                    time_arr, (1, time_arr.shape[0], ntemp_infer, nchannel_infer)
                )
                for time_arr in timeseries_arr
            ]
        )
        if bandidx:
            final = final[:, :, :, np.array(bandidx) - 1]
        if timeidx:
            final = final[:, :, np.array(timeidx) - 1, :]

    else:
        final = tile_stack(
            [
                torch.reshape(time_arr, (1, time_arr.shape[0], ntemp, nchannel))
                for time_arr in timeseries_arr
            ]
        )

    def ts_normalization(x, m, s):
        x = np.rollaxis(x, 2)  # TxCxS -> SxTxC
        x = (x - m) / s
        return torch.tensor(x[None, :, :, :, None])

    normalized_ts = tile_stack(
        [
            ts_normalization(np.rollaxis(tile.numpy(), 0, 3), means, stds)
            for tile in final
        ]
    )
    model = model.to(device)
    pred_list = []
    for i in normalized_ts:
        sim = torch.ones(i.shape[0], i.shape[1], 1).to(device)
        model.eval()
        with torch.no_grad():
            prediction = model(i.float().to(device), sim)

        pred_out = prediction.argmax(dim=1).cpu()
        pred_list.append(pred_out)

    if convertmap:
        convmap = {int(value): int(key) for key, value in convertmap.items()}
        pixel_num_class_mapping = {
            convmap.get(int(key)): int(value)
            for key, value in pixel_num_class_mapping.items()
        }

    remap_pred_list = [
        torch.tensor(
            [pixel_num_class_mapping.get(int(item.numpy())) for item in tile_pred_list]
        )
        for tile_pred_list in pred_list
    ]
    tile_rshp = tile_stack(
        [torch.reshape(i, (1, 1, tile_height, tile_width)) for i in remap_pred_list]
    )
    return tile_rshp


def pixel_classify_pix2pix_image(model, tiles, device, model_info):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]

    num_channel = model_info.get("n_intput_channel", None)
    if num_channel == None:
        num_channel = model_info.get("n_channel", None)
    num_channel_tar = model_info.get("n_channel", None)

    norm_stats_a = model_info.get("NormalizationStats", None)
    model_info["ExtractBands"] = list(range(tiles.shape[1]))
    img_scaled = scale_batch(tiles, model_info, norm_stats_a)
    img_normed = -1 + 2 * img_scaled

    if img_normed.shape[1] < num_channel:
        cont = []
        for j in range(img_normed.shape[0]):
            tile = img_normed[j, :, :, :]
            last_tile = np.expand_dims(tile[tile.shape[0] - 1, :, :], 0)
            res = abs(num_channel - tile.shape[0])
            for i in range(res):
                tile = np.concatenate((tile, last_tile), axis=0)
            cont.append(tile)
        img_normed = np.stack(cont, axis=0)

    pix2pix_predictions = pix2pix_image(model, img_normed, device)
    norm_stats_b = model_info.get("NormalizationStats_b", None)
    pix2pix_predictions = pix2pix_predictions.detach().cpu().numpy()
    if num_channel_tar == None:
        pix2pix_predictions = torch.tensor(
            rescale_batch(pix2pix_predictions, model_info, norm_stats_b)
        )
    else:
        pix2pix_predictions = torch.tensor(
            rescale_batch(pix2pix_predictions, model_info, norm_stats_b)
        )[:, :num_channel_tar, :, :]

    return pix2pix_predictions


def pixel_classify_wnet_image(model, tiles, device, model_info):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]

    num_band_a = model_info.get("n_band_a", None)
    num_band_b = model_info.get("n_band_b", None)
    num_band_tar = model_info.get("n_band_c", None)

    if num_band_a == 1 and num_band_b == 1:
        tile_a = tiles[:, num_band_a, :, :][:, None, :, :]
        tile_b = tiles[:, num_band_b, :, :][:, None, :, :]
    else:
        tile_a = tiles[:, :num_band_a, :, :]
        tile_b = tiles[:, num_band_a:, :, :]

    num_chanel = model_info.get("n_channel", None)

    norm_stats = model_info.get("NormalizationStats", None)
    norm_stats_b = model_info.get("NormalizationStats_b", None)

    img_scaled_a = scale_batch(tile_a, model_info, norm_stats)
    img_normed_a = -1 + 2 * img_scaled_a

    img_scaled_b = scale_batch(tile_b, model_info, norm_stats_b)
    img_normed_b = -1 + 2 * img_scaled_b

    if img_normed_a.shape[1] < num_chanel:
        cont = []
        for j in range(img_normed_a.shape[0]):
            tile = img_normed_a[j, :, :, :]
            last_tile = np.expand_dims(tile[tile.shape[0] - 1, :, :], 0)
            res = abs(num_chanel - tile.shape[0])
            for i in range(res):
                tile = np.concatenate((tile, last_tile), axis=0)
            cont.append(tile)
        img_normed_a = np.stack(cont, axis=0)

    if img_normed_b.shape[1] < num_chanel:
        cont = []
        for j in range(img_normed_b.shape[0]):
            tile = img_normed_b[j, :, :, :]
            last_tile = np.expand_dims(tile[tile.shape[0] - 1, :, :], 0)
            res = abs(num_chanel - tile.shape[0])
            for i in range(res):
                tile = np.concatenate((tile, last_tile), axis=0)
            cont.append(tile)
        img_normed_b = np.stack(cont, axis=0)

    wnet_predictions = wnet_cgan_image(model, img_normed_a, img_normed_b, device)
    norm_stats_c = model_info.get("NormalizationStats_c", None)
    wnet_predictions = wnet_predictions.detach().cpu().numpy()
    wnet_predictions = torch.tensor(
        rescale_batch(wnet_predictions, model_info, norm_stats_c)
    )[:, :num_band_tar, :, :]
    return wnet_predictions


def variable_tile_size_check(json_info, parameters):
    if json_info.get("SupportsVariableTileSize", False):
        parameters.extend(
            [
                {
                    "name": "tile_size",
                    "dataType": "numeric",
                    "value": int(json_info["ImageHeight"]),
                    "required": False,
                    "displayName": "Tile Size",
                    "description": "Tile size used for inferencing",
                }
            ]
        )
    return parameters


def detect_change(model, batch, device, model_info):
    mean = 255 * np.array(
        [0.5] * (len(model_info["ExtractBands"]) // 2), dtype=np.float32
    )
    std = 255 * np.array(
        [0.5] * (len(model_info["ExtractBands"]) // 2), dtype=np.float32
    )
    norm = lambda x: (x - mean) / std
    B, C, H, W = batch.shape
    batch_before = batch[:, : C // 2]
    batch_after = batch[:, C // 2 :]

    if "NormalizationStats" in model_info:
        mean = np.array(
            [0.5] * (len(model_info["ExtractBands"]) // 2), dtype=np.float32
        )
        std = np.array([0.5] * (len(model_info["ExtractBands"]) // 2), dtype=np.float32)
        batch_before = scale_batch(batch_before, model_info, break_extract_bands=True)
        batch_after = scale_batch(batch_after, model_info, break_extract_bands=True)

    batch_before = norm(batch_before.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)
    batch_after = norm(batch_after.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)
    batch_before = torch.tensor(batch_before, device=device).float()
    batch_after = torch.tensor(batch_after, device=device).float()
    from ..._utils.change_detection_data import post_process

    with torch.no_grad():
        predictions = post_process(model(batch_before, batch_after))
    # find the non zero class of the two classes.
    change_class = [c["Value"] for c in model_info["Classes"] if c["Value"] != 0][0]
    predictions[predictions != 0] = change_class
    return predictions[:, 0]


def pixel_classify_pix2pix_hd_image(model, tiles, device, model_info):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]
    num_channel = model_info.get("n_intput_channel", None)
    if num_channel == None:
        num_channel = model_info.get("n_channel", None)
    num_channel_tar = model_info.get("n_channel", None)
    label_nc = model_info.get("label_nc", 0)
    if label_nc == 0:
        norm_stats = model_info.get("NormalizationStats", None)
        img_scaled = scale_batch(tiles, model_info, norm_stats)
        img_normed = -1 + 2 * img_scaled
        if img_normed.shape[1] < num_channel:
            cont = []
            for j in range(img_normed.shape[0]):
                tile = img_normed[j, :, :, :]
                last_tile = np.expand_dims(tile[tile.shape[0] - 1, :, :], 0)
                res = abs(num_channel - tile.shape[0])
                for i in range(res):
                    tile = np.concatenate((tile, last_tile), axis=0)
                cont.append(tile)
            img_normed = np.stack(cont, axis=0)

        pix2pix_predictions = pix2pix_image(model, img_normed, device, label_nc)
        norm_stats_b = model_info.get("NormalizationStats_b", None)
        pix2pix_predictions = pix2pix_predictions.detach().cpu().numpy()
        if num_channel_tar == None:
            pix2pix_predictions = torch.tensor(
                rescale_batch(pix2pix_predictions, model_info, norm_stats_b)
            )
        else:
            pix2pix_predictions = torch.tensor(
                rescale_batch(pix2pix_predictions, model_info, norm_stats_b)
            )[:, :num_channel_tar, :, :]
        return pix2pix_predictions
    else:
        mask_map = model_info.get("mask_map", None)
        for i, j in enumerate(mask_map):
            tiles[tiles == j] = i
        img_normed = tiles

        pix2pix_predictions = pix2pix_image(model, img_normed, device, label_nc)
        pix2pix_predictions = pix2pix_predictions / 2 + 0.5
        return pix2pix_predictions


# functions for test time augmentation and smooth blending for pixel classification


def dihedral_transform(x, k):  # expects [C, H, W]
    flips = []
    if k & 1:
        flips.append(1)
    if k & 2:
        flips.append(2)
    if flips:
        x = torch.flip(x, flips)
    if k & 4:
        x = x.transpose(1, 2)
    return x.contiguous()


def create_interpolation_mask(side, border, device, window_fn="bartlett"):
    if window_fn == "bartlett":
        window = torch.bartlett_window(side, device=device).unsqueeze(0)
        interpolation_mask = window * window.T
    elif window_fn == "hann":
        window = torch.hann_window(side, device=device).unsqueeze(0)
        interpolation_mask = window * window.T
    else:
        linear_mask = torch.linspace(0, 1, border, device=device).repeat(side, 1)
        remainder_tile = torch.ones((side, side - border), device=device)
        interp_tile = torch.cat([linear_mask, remainder_tile], dim=1)

        interpolation_mask = torch.ones((side, side), device=device)
        for i in range(4):
            interpolation_mask = interpolation_mask * interp_tile.rot90(i)
    return interpolation_mask


def unfold_tensor(tensor, tile_size, stride):  # expects tensor  [1, C, H, W]
    mask = torch.ones_like(tensor[0][0].unsqueeze(0).unsqueeze(0))

    unfold = nn.Unfold(kernel_size=(tile_size, tile_size), stride=stride)
    # Apply to mask and original image
    mask_p = unfold(mask)
    patches = unfold(tensor)

    patches = patches.reshape(tensor.size(1), tile_size, tile_size, -1).permute(
        3, 0, 1, 2
    )
    masks = mask_p.reshape(1, tile_size, tile_size, -1).permute(3, 0, 1, 2)
    return masks, (tensor.size(2), tensor.size(3)), patches


def fold_tensor(input_tensor, masks, t_size, tile_size, stride):
    input_tensor_permuted = (
        input_tensor.permute(1, 2, 3, 0).reshape(-1, input_tensor.size(0)).unsqueeze(0)
    )
    mask_tt = masks.permute(1, 2, 3, 0).reshape(-1, masks.size(0)).unsqueeze(0)

    fold = nn.Fold(
        output_size=(t_size[0], t_size[1]),
        kernel_size=(tile_size, tile_size),
        stride=stride,
    )
    output_tensor = fold(input_tensor_permuted) / fold(mask_tt)
    return output_tensor


def split_predict_interpolate(child_image_classifier, normalized_image_tensor):
    kernel_size = child_image_classifier.tytx
    stride = kernel_size - (2 * child_image_classifier.padding)

    # Split image into overlapping tiles
    masks, t_size, patches = unfold_tensor(normalized_image_tensor, kernel_size, stride)

    with torch.no_grad():
        output = child_image_classifier.model(patches)

    interpolation_mask = create_interpolation_mask(
        kernel_size, 0, child_image_classifier.device, "hann"
    )
    output = output * interpolation_mask
    masks = masks * interpolation_mask

    # merge predictions from overlapping chips
    int_surface = fold_tensor(output, masks, t_size, kernel_size, stride)

    return int_surface


def tta_predict(child_image_classifier, normalized_image_tensor, test_time_aug=True):
    all_activations = []

    transforms = [0]
    if test_time_aug:
        if child_image_classifier.json_info["ImageSpaceUsed"] == "MAP_SPACE":
            transforms = list(range(8))
        else:
            transforms = [
                0,
                2,
            ]  # no vertical flips for pixel space (oriented imagery)

    for k in transforms:
        flipped_image_tensor = dihedral_transform(normalized_image_tensor[0], k)
        int_surface = split_predict_interpolate(
            child_image_classifier, flipped_image_tensor.unsqueeze(0)
        )
        corrected_activation = dihedral_transform(int_surface[0], k)

        if k in [5, 6]:
            corrected_activation = dihedral_transform(int_surface[0], k).rot90(
                2, [1, 2]
            )

        all_activations.append(corrected_activation)

    all_activations = torch.stack(all_activations)

    return all_activations


def update_pixels_tta(
    child_image_classifier, tlc, shape, props, **pixelBlocks
):  # 8 x 224 x 224 x 3
    model_info = child_image_classifier.json_info

    class_values = [clas["Value"] for clas in model_info["Classes"]]
    is_contiguous = is_cont([0] + class_values)

    if not is_contiguous:
        pixel_mapping = [0] + class_values
        idx2pixel = {i: d for i, d in enumerate(pixel_mapping)}

    input_image = pixelBlocks["raster_pixels"].astype(np.float32)
    input_image_tensor = (
        torch.tensor(input_image).to(child_image_classifier.device).float()
    )

    if "NormalizationStats" in model_info:
        normalized_image_tensor = normalize_batch(input_image_tensor.cpu(), model_info)
        normalized_image_tensor = normalized_image_tensor.float().to(
            input_image_tensor.device
        )
    else:
        from torchvision import transforms

        normalize = transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        normalized_image_tensor = normalize(input_image_tensor / 255.0).unsqueeze(0)

    all_activations = tta_predict(
        child_image_classifier,
        normalized_image_tensor,
        test_time_aug=child_image_classifier.use_tta,
    )
    # probability of each class in 2nd dimension
    all_activations = all_activations.mean(dim=0, keepdim=True)
    softmax_surface = all_activations.softmax(dim=1)

    ignore_mapped_class = model_info.get("ignore_mapped_class", [])

    for k in ignore_mapped_class:
        softmax_surface[:, k] = -1

    if not child_image_classifier.predict_background:
        softmax_surface[:, 0] = -1

    result = softmax_surface.max(dim=1)[1]

    if not is_contiguous:
        result = remap(result, idx2pixel)

    pad = child_image_classifier.padding

    return result.cpu().numpy().astype("i4")[:, pad : -pad or None, pad : -pad or None]


def update_pixels_img_trans(self, tlc, shape, props, **pixelBlocks):
    kernel_size = self.json_info["ImageHeight"]
    stride = kernel_size - (2 * self.padding)

    model_name = self.json_info.get("ModelName")

    pixelblock = pixelBlocks["raster_pixels"].astype(np.float32)
    pixelblock_image_tensor = tensor(pixelblock).float().unsqueeze(0).cpu()

    masks, t_size, patches = unfold_tensor(pixelblock_image_tensor, kernel_size, stride)

    if model_name == "Pix2PixHD":
        prediction = pixel_classify_pix2pix_hd_image(
            self.model, patches, self.device, model_info=self.json_info
        )
    elif model_name == "CycleGAN":
        prediction = pixel_classify_cyclegan_image(
            self.model, patches, self.device, self.direction, model_info=self.json_info
        )
    elif model_name == "Pix2Pix":
        prediction = pixel_classify_pix2pix_image(
            self.model, patches, self.device, model_info=self.json_info
        )
    elif model_name == "SuperResolution":
        prediction = pixel_classify_superres_image(
            self.model,
            patches,
            self.device,
            self.json_info,
            getattr(self, "sampling", None),
            getattr(self, "n_timestep", None),
        )
    elif model_name == "WNetcGAN":
        prediction = pixel_classify_wnet_image(
            self.model, patches, self.device, model_info=self.json_info
        )

    interpolation_mask = create_interpolation_mask(kernel_size, 0, self.device, "hann")

    output = (tensor(prediction).to(self.device)) * interpolation_mask
    masks_inp = (tensor(masks).to(self.device)) * interpolation_mask

    merged_preds = fold_tensor(output, masks_inp, t_size, kernel_size, stride)

    return merged_preds.cpu().numpy()[
        0,
        :,
        self.padding : -self.padding or None,
        self.padding : -self.padding or None,
    ]



"""
Contains the base class that all server object inherit from.
"""
from __future__ import annotations
from urllib.request import HTTPError
from arcgis.gis import GIS
from arcgis._impl.common._isd import InsensitiveDict
from typing import Dict, Any, Optional, List


###########################################################################
class KubeSecurityCert(object):
    """
    The certificates resource provides access to child operations and
    resources that can be used to manage all the security certificates
    configured with an organization.
    """

    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None

    def __init__(self, url: str, gis: GIS) -> "KubeSecurityIngress":
        """class initializer"""
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = InsensitiveDict(result)
            else:
                self._json_dict = {}
                self._properties = InsensitiveDict({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = InsensitiveDict({})

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> InsensitiveDict:
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def url(self) -> str:
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    @url.setter
    def url(self, value: str):
        """gets/sets the service url"""
        self._url = value
        self._refresh()

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def identity_certs(self) -> list:
        """
        Lists all the certificates currently configured with the organization

        :return: List
        """
        url = self._url + "/identity"
        params = {"f": "json"}
        return self._con.get(url, params).get("certificates", [])

    # ----------------------------------------------------------------------
    def remove_identity_cert(self, cert_id: str) -> bool:
        """Deletes an Identity Certificate by ID

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        cert_id                Required String. The unique identifier of the certificate.
        ==================     ====================================================================

        """
        url = self._url + f"/identity/{cert_id}/delete"
        params = {"f": "json"}
        return self._con.post(url, params).get("status", "failed") == "success"

    # ----------------------------------------------------------------------
    def load_identity_cert(self, pfx: str, password: str, name: str) -> bool:
        """
        Imports an existing identity certificate in PKCS #12 (.pfx) format
        into the keystore. An imported certificate can be assigned to the
        Ingress controller by setting the certificate name property via the
        update operation.

        :return: bool
        """
        params = {
            "f": "json",
            "certificateName": name,
            "certificatePassword": password,
        }
        files = {"certificatePfxFile": pfx}
        url = self._url + "/identity/import"
        ret = self._gis._con.post(path=url, postdata=params, files=files)
        return ret.get("status", "false") == "success"

    # ----------------------------------------------------------------------
    @property
    def trust_certs(self) -> list:
        """
        Lists all the trust certificates configured with the organization
        :return: list
        """
        params = {"f": "json"}

        url = self._url + "/trust"
        return self._con.get(url, params).get("certificates", [])

    # ----------------------------------------------------------------------
    def load_trust_cert(self, cert: str, name: str) -> bool:
        """
        This operation imports a trust certificate, in either PEM
        (.cer or .crt files) or a binary (.der) format. Once a trust
        certificate is imported, the corresponding pods that will use the
        certificate are automatically restarted.

        :return: bool
        """
        params = {"f": "json", "certificateName": name}
        files = {"trustCertificateFile": cert}
        url = self._url + "/trust/import"
        ret = self._gis._con.post(path=url, postdata=params, files=files)
        return ret.get("status", "false") == "success"

    def get_cert(self, cert_type: str, cert_id: str) -> dict:
        """
        Obtains a single certificate for a given type and ID

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        cert_type              Required String. The type of certificate to search for. This can be 'trust' or 'identity'.
        ------------------     --------------------------------------------------------------------
        cert_id                Required String. The unique identifier of the certificate.
        ==================     ====================================================================

        :return: Dict
        """
        if cert_type.lower() == "trust":
            url = self._url + f"/trust/{cert_id}"
            params = {"f": "json"}
            return self._con.get(url, params)
        elif cert_type.lower() == "identity":
            url = self._url + f"/identity{cert_id}"
            params = {"f": "json"}
            return self._con.get(url, params)
        else:
            raise ValueError("Invalid certificate type.")

    # ----------------------------------------------------------------------
    def remove_trust_cert(self, cert_id: str) -> bool:
        """Deletes an Identity Certificate by ID

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        cert_id                Required String. The unique identifier of the certificate.
        ==================     ====================================================================

        """
        url = self._url + f"/trust/{cert_id}/delete"
        params = {"f": "json"}
        return self._con.post(url, params).get("status", "failed") == "success"


###########################################################################
class KubeSecuritySAML(object):
    """
    The saml resource returns information about the SAML configuration for
    an organization. If SAML is configured, the enabled property will
    return as true and `identityCertificateName` will show the name of the
    imported identity certificate.
    """

    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None

    def __init__(self, url: str, gis: GIS) -> "KubeSecurityIngress":
        """class initializer"""
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = InsensitiveDict(result)
            else:
                self._json_dict = {}
                self._properties = InsensitiveDict({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = InsensitiveDict({})

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> InsensitiveDict:
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def url(self) -> str:
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    @url.setter
    def url(self, value: str):
        """gets/sets the service url"""
        self._url = value
        self._refresh()

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def settings(self) -> dict:
        """
        This get/sets the SAML certification information.

        :return: dict

        """
        return dict(self.properties)

    # ----------------------------------------------------------------------
    @settings.setter
    def settings(self, value: dict):
        """
        This get/sets the SAML certification information.

        :return: dict
        """
        url = self.url + "/update"
        params = {"f": "json", "samlSecurityConfig": value}
        res = self._con.post(url, params)
        if res.get("status", "failed") == "success":
            self._refresh()


###########################################################################
class KubeSecurityIngress(object):
    """
    The ingress resource returns the currently configured security
    information for the Ingress controller. You can update ingress security
    configuration properties using the update operation. The update
    operation must be used when adding an imported wildcard certificate for
    the Ingress controller.
    """

    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None

    def __init__(self, url: str, gis: GIS) -> "KubeSecurityIngress":
        """class initializer"""
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = InsensitiveDict(result)
            else:
                self._json_dict = {}
                self._properties = InsensitiveDict({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = InsensitiveDict({})

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> InsensitiveDict:
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def url(self) -> str:
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    @url.setter
    def url(self, value: str):
        """gets/sets the service url"""
        self._url = value
        self._refresh()

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def settings(self) -> dict:
        """
        gets/sets the ingress configuration properties

        :return: dict

        """
        return dict(self.properties)

    # ----------------------------------------------------------------------
    @settings.setter
    def settings(self, value: dict):
        """
        gets/sets the ingress configuration properties

        :return: dict
        """
        url = self.url + "/update"
        params = {"f": "json", "samlSecurityConfig": value}
        res = self._con.post(url, params)
        if res.get("status", "failed") == "success":
            self._refresh()


###########################################################################
class KubeSecurityConfig(object):
    """
    Allows the user to manage the security configuration for an ArcGIS Enterprise for Kubernetes deployment.
    """

    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None

    def __init__(self, url: str, gis: GIS) -> "KubeSecurityConfig":
        """class initializer"""
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = InsensitiveDict(result)
            else:
                self._json_dict = {}
                self._properties = InsensitiveDict({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = InsensitiveDict({})

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> InsensitiveDict:
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def url(self) -> str:
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    @url.setter
    def url(self, value: str):
        """gets/sets the service url"""
        self._url = value
        self._refresh()

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def settings(self) -> dict:
        """gets/sets the current secutiry settings for the deployment"""
        return dict(self.properties)

    # ----------------------------------------------------------------------
    @settings.setter
    def settings(self, value: dict):
        """gets/sets the current secutiry settings for the deployment"""
        url = self.url + "/update"
        params = {"f": "json", "securityConfig": value}
        res = self._con.post(url, params)
        if res.get("status", "failed") == "success":
            self._refresh()

    # ----------------------------------------------------------------------
    def test(self, user_store: dict = None, role_store: dict = None) -> bool:
        """
        Users can test the connection to a user or role (group) store.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        user_store             Optional dict. Specifies the user store properties. This parameter
                               accepts as input all the properties as defined in the
                               `user_store` and `role_store` section of the Kubernetes help
                               documentation.
        ------------------     --------------------------------------------------------------------
        role_store             Optional dict. pecifies the role (group) store properties. This parameter
                               accepts as input all the properties as defined in the
                               `user_store` and `role_store` section of the Kubernetes help
                               documentation.
        ==================     ====================================================================

        :return: boolean
        """
        url = f"{self._url}/testIdentityStore"
        params = {
            "f": "json",
        }
        if user_store:
            params["userStoreConfig"] = user_store
        if role_store:
            params["roleStoreConfig"] = role_store
        return self._con.post(url, params).get("status", "fail") == "success"

    # ----------------------------------------------------------------------
    def update_stores(self, user_store: dict = None, role_store: dict = None) -> bool:
        """
        Users can modify the user or role (group) identity stores.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        user_store             Optional dict. Specifies the user store properties. This parameter accepts as input all the properties as defined in the userStoreConfig and roleStoreConfig section of the Kubernetes help doctumentation.
        ------------------     --------------------------------------------------------------------
        role_store             Optional dict. pecifies the role (group) store properties. This parameter accepts as input all the properties as defined in the ArcGIS for Kubernetes help doctumentation.
        ==================     ====================================================================

        :return: boolean
        """
        url = f"{self._url}/updateIdentityStore"
        params = {
            "f": "json",
        }
        if user_store:
            params["userStoreConfig"] = user_store
        if role_store:
            params["roleStoreConfig"] = role_store
        return self._con.post(url, params).get("status", "fail") == "success"


###########################################################################
class KubeSecurity(object):
    """Allows users to configure the Security settings on the kubernetes infrastructure"""

    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None
    _config = None
    _ingress = None
    _saml = None
    _certs = None

    def __init__(self, url: str, gis: GIS) -> "KubeSecurity":
        """class initializer"""
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = InsensitiveDict(result)
            else:
                self._json_dict = {}
                self._properties = InsensitiveDict({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = InsensitiveDict({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def url(self):
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    @url.setter
    def url(self, value):
        """gets/sets the service url"""
        self._url = value
        self._refresh()

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def configuration(self) -> "KubeSecurityConfig":
        """Returns the currently active security configuration for an ArcGIS Enterprise for Kubernetes deployment"""
        if self._config is None:
            url = self._url + "/config"
            self._config = KubeSecurityConfig(url=url, gis=self._gis)
        return self._config

    # ----------------------------------------------------------------------
    @property
    def ingress(self) -> "KubeSecurityIngress":
        """Returns a manager to configure the ingress settings.

        :return:
            A KubeSecurityIngress object.
        """
        if self._ingress is None:
            url = self._url + "/ingress"
            self._ingress = KubeSecurityIngress(url, gis=self._gis)
        return self._ingress

    # ----------------------------------------------------------------------
    @property
    def saml(self) -> "KubeSecuritySAML":
        """
        Returns a manager to work with the SAML settings for the organization

        :return: KubeSecuritySAML
        """
        if self._saml is None:
            url = self._url + "/saml"
            self._saml = KubeSecuritySAML(url, gis=self._gis)
        return self._saml

    # ----------------------------------------------------------------------
    @property
    def certificates(self) -> "KubeSecurityCert":
        """
        Provides access to the certificate manager for the Kubernetes infrastructure

        :returns:
            A KubeSecurityCert object.

        """
        if self._certs is None:
            url = self._url + "/certificates"
            self._certs = KubeSecurityCert(url, gis=self._gis)
        return self._certs



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


_base_ = "./deeplabv3.py"
model = dict(
    pretrained="open-mmlab://resnest101",
    backbone=dict(
        type="ResNeSt",
        stem_channels=128,
        radix=2,
        reduction_factor=4,
        avg_down_stride=True,
    ),
)

checkpoint = "https://download.openmmlab.com/mmsegmentation/v0.5/resnest/deeplabv3_s101-d8_512x1024_80k_cityscapes/deeplabv3_s101-d8_512x1024_80k_cityscapes_20200807_144429-b73c4270.pth"



from fastai.vision import ImageSegment, Image
from fastai.vision.image import open_image, show_image, pil2tensor
from fastai.vision.data import SegmentationProcessor, ImageList
from fastai.layers import CrossEntropyFlat
from fastai.basic_train import LearnerCallback
from fastai.core import is_listy
from .._utils.common import (
    ArcGISMSImage,
    get_top_padding,
    kwarg_fill_none,
    find_data_loader,
    get_nbatches,
    dynamic_range_adjustment,
    image_tensor_checks_plotting,
    get_symbology_bands,
    predict_batch,
    denorm_x,
    get_nbatches,
    GDAL_INSTALL_MESSAGE,
    image_batch_stretcher,
)
from .._utils.env import HAS_GDAL
from .._utils.pixel_classification import analyze_pred_pixel_classification
from .._utils.env import is_arcgispronotebook
from .._utils.utils import check_imbalance
import torch
import warnings
import PIL
import numpy as np
import os
import math
from typing import Callable
import warnings


def _class_array_to_rbg(ca: "classified_array", cm: "color_mapping", nodata=0):
    im = np.expand_dims(ca, axis=2).repeat(3, axis=2)
    white_mask = im[im == nodata]
    for x in np.unique(im):
        if not x == nodata:
            for i in range(3):
                im[:, :, i][im[:, :, i] == x] = cm[x][i]
    im[white_mask] = 255
    return im


# def _show_batch_unet_multispectral(self, nrows=3, ncols=3, n_items=None, index=0, rgb_bands=None, nodata=0, alpha=0.7, imsize=5): # Proposed Parameters
def _show_batch_unet_multispectral(
    self, rows=3, alpha=0.7, **kwargs
):  # parameters adjusted in kwargs
    import matplotlib.pyplot as plt
    from .._data import _tensor_scaler

    nrows = rows
    ncols = 3
    if kwargs.get("ncols", None) is not None:
        ncols = kwargs.get("ncols")

    n_items = None
    if kwargs.get("n_items", None) is not None:
        n_items = kwargs.get("n_items")

    type_data_loader = kwargs.get(
        "data_loader", "training"
    )  # options : traininig, validation, testing
    if type_data_loader == "training":
        data_loader = self.train_dl
    elif type_data_loader == "validation":
        data_loader = self.valid_dl
    elif type_data_loader == "testing":
        data_loader = self.test_dl
    else:
        e = Exception(
            f"could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing "
        )
        raise (e)

    rgb_bands = self._symbology_rgb_bands
    if kwargs.get("rgb_bands", None) is not None:
        rgb_bands = kwargs.get("rgb_bands")

    nodata = 0
    if kwargs.get("nodata", None) is not None:
        nodata = kwargs.get("nodata")

    index = 0
    if kwargs.get("index", None) is not None:
        index = kwargs.get("index")

    imsize = 5
    if kwargs.get("imsize", None) is not None:
        imsize = kwargs.get("imsize")

    statistics_type = kwargs.get(
        "statistics_type", "dataset"
    )  # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get(
        "stretch_type", "minmax"
    )  # Accepted Values `minmax`, `percentclip`

    e = Exception(
        "`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1."
    )
    symbology_bands = []
    if not (len(rgb_bands) == 3 or len(rgb_bands) == 1):
        raise (e)
    for b in rgb_bands:
        if type(b) == str:
            b_index = self._bands.index(b)
        elif type(b) == int:
            self._bands[
                b
            ]  # To check if the band index specified by the user really exists.
            b_index = b
        else:
            raise (e)
        b_index = self._extract_bands.index(b_index)
        symbology_bands.append(b_index)

    # Get Batch
    if n_items is None:
        n_items = nrows * ncols
    else:
        nrows = math.ceil(n_items / ncols)
    n_items = min(n_items, len(self.x))

    x_batch, y_batch = get_nbatches(data_loader, n_items)
    x_batch = torch.cat(x_batch)
    y_batch = torch.cat(y_batch)
    # Denormalize X
    x_batch = (
        self._scaled_std_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch)
        * x_batch
    ) + self._scaled_mean_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch)

    # Extract RGB Bands
    symbology_x_batch = x_batch[:, symbology_bands]
    if stretch_type is not None:
        symbology_x_batch = image_batch_stretcher(
            symbology_x_batch, stretch_type, statistics_type
        )

    # Channel first to channel last and clamp float values to range 0 - 1 for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Get color Array
    color_array = self._multispectral_color_array
    color_array[1:, 3] = alpha

    # Size for plotting
    fig, axs = plt.subplots(
        nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
    )
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            if nrows == 1 and ncols == 1:
                axi = axs
            else:
                axi = axs[r][c]
            if idx < symbology_x_batch.shape[0]:
                axi.imshow(symbology_x_batch[idx])
                y_rgb = color_array[y_batch[idx][0]]  # .cpu().numpy()
                # y_rgb = _class_array_to_rbg(y_batch[idx][0], self._multispectral_color_mapping, nodata)
                axi.imshow(y_rgb, alpha=alpha)
                axi.axis("off")
            else:
                axi.axis("off")
            idx += 1
    #


class ArcGISImageSegment(Image):
    "Support applying transforms to segmentation masks data in `px`."

    def __init__(self, x, color_mapping=None):
        super(ArcGISImageSegment, self).__init__(x)
        self.color_mapping = color_mapping

    def lighting(self, func, *args, **kwargs):
        return self

    def refresh(self):
        self.sample_kwargs["mode"] = "nearest"
        return super().refresh()

    @property
    def data(self):
        "Return this image pixels as a `LongTensor`."
        return self.px.long()

    def show(
        self,
        ax=None,
        figsize: tuple = (3, 3),
        title=None,
        hide_axis: bool = True,
        cmap=None,
        alpha: float = 0.5,
        **kwargs,
    ):
        "Show the `ImageSegment` on `ax`."
        if is_no_color(self.color_mapping):
            ## This condition will not be true.
            ax = show_image(
                self,
                ax=ax,
                hide_axis=hide_axis,
                cmap="tab20",
                figsize=figsize,
                interpolation="nearest",
                alpha=alpha,
                vmin=0,
                **kwargs,
            )
        else:
            color_mapping = torch.tensor(list(self.color_mapping.values()))
            color_mapping = torch.cat(
                (
                    color_mapping.float() / 255,
                    torch.tensor([float(alpha)] * len(color_mapping)).view(-1, 1),
                ),
                dim=1,
            )
            color_mapping = torch.cat(
                (torch.tensor([0.0, 0.0, 0.0, 0.0]).view(1, -1), color_mapping), dim=0
            )
            try:
                color_im = color_mapping[self.data[0]].permute(2, 0, 1)
            except IndexError as e:
                if HAS_GDAL:
                    message = f"Encountered invalid values in training label values, please check your training data."
                else:
                    message = (
                        f"Encountered invalid values while reading training labels. Please install gdal for better support.\n\n"
                        + GDAL_INSTALL_MESSAGE
                    )

                raise Exception(f"{e} \n\n{message}")

            ax = show_image(
                color_im,
                ax=ax,
                hide_axis=hide_axis,
                cmap=cmap,
                figsize=figsize,
                interpolation="nearest",
                alpha=alpha,
                vmin=0,
                **kwargs,
            )
        if title:
            ax.set_title(title)


class ArcGISMultispectralImageSegment:
    def __init__(self, tensor):
        self.data = tensor
        self.size = tensor.shape
        self.shape = tensor.shape


def is_no_color(color_mapping):
    if isinstance(color_mapping, dict):
        color_mapping = list(color_mapping.values())
    return (np.array(color_mapping) == [-1.0, -1.0, -1.0]).any()


def is_contiguous(class_values):
    flag = True
    for i in range(len(class_values) - 1):
        if class_values[i] + 1 != class_values[i + 1]:
            flag = False
    return flag


def map_to_contiguous(tensor, mapping):
    modified_tensor = torch.zeros_like(tensor)
    for i, value in enumerate(mapping):
        modified_tensor[tensor == value] = i
    return modified_tensor


class ArcGISSegmentationLabelList(ImageList):
    "`ItemList` for segmentation masks."
    _processor = SegmentationProcessor

    def __init__(
        self, items, classes=None, class_mapping=None, color_mapping=None, **kwargs
    ):
        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append("classes")
        self.classes, self.loss_func = classes, CrossEntropyFlat(axis=1)
        self.is_contiguous = is_contiguous(
            sorted([0] + list(self.class_mapping.keys()))
        )
        if not self.is_contiguous:
            self.pixel_mapping = [0] + list(self.class_mapping.keys())

    def _open_rgb(self, fn):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)  # EXIF warning from TiffPlugin
            x = PIL.Image.open(fn)
            if x.palette is not None:
                x = x.convert("P")
            else:
                x = x.convert("L")
            x = pil2tensor(x, np.float32)

        if not self.is_contiguous:
            x = map_to_contiguous(x, self.pixel_mapping)

        return ArcGISImageSegment(x, color_mapping=self.color_mapping)

    def analyze_pred(
        self, pred, thresh=0.5, ignore_mapped_class=[], model=None, thinning=None
    ):
        if getattr(model, "_is_model_extension", False):
            if thinning is None:
                pred = model._model_conf.post_process(pred, thresh)
            else:
                pred = model._model_conf.post_process(pred, thresh, thinning)
            return pred

        if is_listy(pred):
            pred = pred[0]

        if ignore_mapped_class == []:
            return pred.argmax(dim=0)[None]
        else:
            for k in ignore_mapped_class:
                pred[k] = -1
            return pred.argmax(dim=0)[None]

    def reconstruct(self, t):
        return ArcGISImageSegment(t, color_mapping=self.color_mapping)

    def open(self, fn):
        x = ArcGISMSImage.open(fn).data
        if not self.is_contiguous:
            x = map_to_contiguous(x, self.pixel_mapping)
        return ArcGISImageSegment(x, color_mapping=self.color_mapping)


class ArcGISSegmentationItemList(ImageList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False
    _div = None
    _imagery_type = None

    def open(self, fn):
        return ArcGISMSImage.open(fn, div=self._div, imagery_type=self._imagery_type)

    def check_class_imbalance(
        self, func: Callable, class_mapping, stratify=False, class_imbalance_pct=0.01
    ):
        try:
            labelval = [(func(o)) for o in self.items]
            total_sample = np.concatenate(labelval)
            total_sample = total_sample[total_sample != "0"]
            unique_sample = set(total_sample)
            imabalanced_class_list = []

            for sample in unique_sample:
                if (total_sample == sample).sum() < len(
                    total_sample
                ) * class_imbalance_pct:
                    imabalanced_class_list.append(class_mapping[int(sample)])

            if stratify == True and len(imabalanced_class_list) > 0:
                warnings.warn(
                    f'We see a class imbalance in the dataset. The class(es) {",".join(imabalanced_class_list)} doesnt have enough data points in your dataset.'
                )
            elif stratify == False and len(imabalanced_class_list) > 0:
                warnings.warn(
                    f'We see a class imbalance in the dataset. The class(es) {",".join(imabalanced_class_list)} doesnt have enough data points in your dataset. Although, class imbalance cannot be overcome easily, adding the parameter stratify = True will to a certain extent help get over this problem.'
                )
        except Exception as e:
            warnings.warn(f"Unable to check for class imbalance [reason : {e}]")

        return self

    def label_list_from_func(self, func: Callable):
        "Apply `func` to every input to get its label."
        import pandas as pd

        self._list_of_labels = ["_".join(func(o)) for o in self.items]
        self._idx_label_tuple_list = [
            (i, label) for i, label in enumerate(self._list_of_labels)
        ]
        label_series = pd.Series(self._list_of_labels)
        single_instance_labels = list(
            label_series.value_counts()[label_series.value_counts() == 1].index
        )
        self._label_idx_mapping = {
            label: i for i, label in enumerate(self._list_of_labels)
        }
        for (
            label
        ) in single_instance_labels:  # adding duplicate instance of unique labels
            self._idx_label_tuple_list.append((self._label_idx_mapping[label], label))
        return self

    def stratified_split_by_pct(self, valid_pct: float = 0.2, seed: int = None):
        try:
            "Split the items in a stratified manner by putting `valid_pct` in the validation set, optional `seed` can be passed."
            from sklearn.model_selection import train_test_split
            import random, math

            if valid_pct == 0.0:
                return self.split_none()
            if seed is not None:
                np.random.seed(seed)
            if (
                len(set(self._list_of_labels)) > len(self._list_of_labels) * valid_pct
            ):  # if validation samples length is less than unique labels
                classes = len(set(self._list_of_labels))
                xlen = len(self._list_of_labels)
                sample_shortage = math.ceil((classes - xlen * valid_pct) / valid_pct)
                print(sample_shortage)
                extra_samples = random.choices(
                    self._idx_label_tuple_list, k=sample_shortage
                )
                self._idx_label_tuple_list.extend(extra_samples)
            X, y = [], []
            for index, label in self._idx_label_tuple_list:
                X.append(index)
                y.append(label)
            train_idx, val_idx, _, _ = train_test_split(
                X, y, test_size=valid_pct, random_state=seed, stratify=y
            )
            return self.split_by_idxs(train_idx, val_idx)
        except Exception as e:
            warnings.warn(
                f"Unable to perform stratified splitting [reason : {e}], falling back to random split"
            )
            return self.split_by_rand_pct(valid_pct=valid_pct, seed=seed)


class ArcGISSegmentationMSLabelList(ArcGISSegmentationLabelList):
    def open(self, fn):
        from osgeo import gdal

        path = str(os.path.abspath(fn))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            x = gdal.Open(path).ReadAsArray()
        x = torch.tensor(x.astype(np.float32))[None]
        if not self.is_contiguous:
            x = map_to_contiguous(x, self.pixel_mapping)
        return ArcGISImageSegment(x, color_mapping=self.color_mapping)


class LabelCallback(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)
        self.label_mapping = {
            value: (idx + 1)
            for idx, value in enumerate(learn.data.class_mapping.keys())
        }

    def on_batch_begin(self, last_input, last_target, **kwargs):
        """
        This callback is not used anymore.
        Not using this callback has increased the unet model training speed.
        """
        # modified_target = torch.zeros_like(last_target)
        # for label, idx in self.label_mapping.items():
        #     modified_target[last_target==label] = idx
        return {"last_input": last_input, "last_target": last_target}


# def show_results_multispectral(self, nrows=3, index=0, type_ds='valid', rgb_bands=None, nodata=0, alpha=0.7, imsize=5, top=0.97): # Proposed Parameters
def show_results_multispectral(
    self, nrows=5, alpha=0.7, **kwargs
):  # parameters adjusted in kwargs
    import matplotlib.pyplot as plt

    # Get Number of items
    ncols = 2
    return_fig = kwargs.get("return_fig", False)
    type_data_loader = kwarg_fill_none(
        kwargs, "data_loader", "validation"
    )  # options : traininig, validation, testing
    data_loader = find_data_loader(type_data_loader, self._data)
    if getattr(self, "name", "") in ["MultiTaskRoadExtractor"]:
        data_loader = find_data_loader(type_data_loader, self._orient_data)
        self._data.batch_size = 1

    nodata = kwarg_fill_none(kwargs, "nodata", 0)

    index = kwarg_fill_none(kwargs, "index", 0)

    imsize = kwarg_fill_none(kwargs, "imsize", 5)

    top = kwargs.get("top", None)
    title_font_size = 16
    if top is None:
        top = get_top_padding(
            title_font_size=title_font_size, nrows=nrows, imsize=imsize
        )

    statistics_type = kwarg_fill_none(
        kwargs, "statistics_type", "dataset"
    )  # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get(
        "stretch_type", "minmax"
    )  # Accepted Values `minmax`, `percentclip`

    # get batches
    x_batch, y_batch = get_nbatches(
        data_loader, math.ceil(nrows / self._data.batch_size)
    )
    symbology_x_batch = x_batch = torch.cat(x_batch)
    if getattr(self, "name", "") in ["MultiTaskRoadExtractor"]:
        y_batch = (
            torch.stack([item for sublist in y_batch for item in sublist[0]])
            .type(torch.long)
            .unsqueeze(1)
        )
    else:
        y_batch = torch.cat(y_batch)

    symbology_bands = [0, 1, 2]
    if self._is_multispectral:
        # Get RGB Bands for plotting
        rgb_bands = kwarg_fill_none(
            kwargs, "rgb_bands", self._data._symbology_rgb_bands
        )

        # Get Symbology bands
        symbology_bands = get_symbology_bands(
            rgb_bands, self._data._extract_bands, self._data._bands
        )

    # Get Predictions
    activation_store = []
    for i in range(0, x_batch.shape[0], self._data.batch_size):
        activations = predict_batch(self, x_batch[i : i + self._data.batch_size])
        activation_store.append(activations)

    # Analyze Pred
    if getattr(self, "name", "") in ["MultiTaskRoadExtractor"]:
        activation_store = [x[0] for x in activation_store]
    predictions = analyze_pred_pixel_classification(self, activation_store)

    # Denormalize X
    x_batch = denorm_x(x_batch, self)

    # Extract RGB Bands for plotting
    symbology_x_batch = x_batch[:, symbology_bands]
    if stretch_type is not None:
        symbology_x_batch = image_batch_stretcher(
            symbology_x_batch, stretch_type, statistics_type
        )

    # Apply Image Strecthing
    if statistics_type == "DRA":
        symbology_x_batch = dynamic_range_adjustment(symbology_x_batch)

    symbology_x_batch = image_tensor_checks_plotting(symbology_x_batch)

    # Get color Array
    color_array = self._data._multispectral_color_array
    color_array[1:, 3] = alpha

    # Size for plotting
    nrows = min(nrows, symbology_x_batch.shape[0])
    fig, axs = plt.subplots(
        nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
    )
    plt.subplots_adjust(top=top)
    fig.suptitle("Ground Truth / Predictions", fontsize=title_font_size)
    for r in range(nrows):
        if nrows == 1:
            axi = axs
        else:
            axi = axs[r]
        if r < symbology_x_batch.shape[0]:
            axi[0].imshow(symbology_x_batch[r].cpu().numpy())
            y_rgb = color_array[y_batch[r][0]]
            axi[0].imshow(y_rgb, alpha=alpha)
            axi[1].imshow(symbology_x_batch[r].cpu().numpy())
            p_rgb = color_array[predictions[r]]
            axi[1].imshow(p_rgb, alpha=alpha)
        axi[0].axis("off")
        axi[1].axis("off")
    #
    if is_arcgispronotebook():
        plt.show()
    if return_fig:
        return fig, axs



# model settings
custom_imports = dict(imports=["arcgis.learn.models._prithvi_archs"])

bands = [1, 2, 3, 8, 11, 12]
# required bands Blue, Green, Red, Narrow NIR, SWIR 1, SWIR 2

loss_weights_multi = [
    0.386375,
    0.661126,
    0.548184,
    0.640482,
    0.876862,
    0.925186,
    3.249462,
    1.542289,
    2.175141,
    2.272419,
    3.062762,
    3.626097,
    1.198702,
]

img_norm_crop_model = dict(
    means=[
        494.905781,
        815.239594,
        924.335066,
        2968.881459,
        2634.621962,
        1739.579917,
        494.905781,
        815.239594,
        924.335066,
        2968.881459,
        2634.621962,
        1739.579917,
        494.905781,
        815.239594,
        924.335066,
        2968.881459,
        2634.621962,
        1739.579917,
    ],
    stds=[
        284.925432,
        357.84876,
        575.566823,
        896.601013,
        951.900334,
        921.407808,
        284.925432,
        357.84876,
        575.566823,
        896.601013,
        951.900334,
        921.407808,
        284.925432,
        357.84876,
        575.566823,
        896.601013,
        951.900334,
        921.407808,
    ],
)


CLASSES = (
    "Natural Vegetation",
    "Forest",
    "Corn",
    "Soybeans",
    "Wetlands",
    "Developed/Barren",
    "Open Water",
    "Winter Wheat",
    "Alfalfa",
    "Fallow/Idle Cropland",
    "Cotton",
    "Sorghum",
    "Others",
    "nodata",
)

norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="TemporalEncoderDecoder",
    pretrained=None,
    frozen_backbone=True,
    backbone=dict(
        type="TemporalViTEncoder",
        pretrained=None,
        img_size=224,
        patch_size=16,
        num_frames=3,
        tubelet_size=1,
        in_chans=len(bands),
        embed_dim=768,
        depth=6,
        num_heads=8,
        mlp_ratio=4.0,
        norm_pix_loss=False,
    ),
    neck=dict(
        type="ConvTransformerTokensToEmbeddingNeck",
        embed_dim=768 * 3,
        output_embed_dim=768 * 3,
        drop_cls_token=True,
        Hp=14,
        Wp=14,
    ),
    decode_head=dict(
        in_channels=768 * 3,
        type="FCNHead",
        in_index=-1,
        channels=256,
        num_convs=1,
        concat_input=False,
        dropout_ratio=0.1,
        norm_cfg=dict(type="BN", requires_grad=True),
        align_corners=False,
        loss_decode=dict(
            type="CrossEntropyLoss",
            use_sigmoid=False,
            class_weight=loss_weights_multi,
            avg_non_ignore=True,
        ),
    ),
    auxiliary_head=dict(
        in_channels=768 * 3,
        type="FCNHead",
        in_index=-1,
        channels=256,
        num_convs=2,
        concat_input=False,
        dropout_ratio=0.1,
        norm_cfg=dict(type="BN", requires_grad=True),
        align_corners=False,
        loss_decode=dict(
            type="CrossEntropyLoss",
            use_sigmoid=False,
            class_weight=loss_weights_multi,
            avg_non_ignore=True,
        ),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode="whole"),
)

checkpoint = "https://huggingface.co/ibm-nasa-geospatial/Prithvi-100M-multi-temporal-crop-classification/resolve/main/multi_temporal_crop_classification_Prithvi_100M.pth"



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# box AP=38.9
_base_ = "./_base_/models/faster_rcnn_r50_fpn.py"
model = dict(
    roi_head=dict(
        type="DynamicRoIHead",
        bbox_head=dict(
            type="Shared2FCBBoxHead",
            in_channels=256,
            fc_out_channels=1024,
            roi_feat_size=7,
            num_classes=80,
            bbox_coder=dict(
                type="DeltaXYWHBBoxCoder",
                target_means=[0.0, 0.0, 0.0, 0.0],
                target_stds=[0.1, 0.1, 0.2, 0.2],
            ),
            reg_class_agnostic=False,
            loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
            loss_bbox=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1.0),
        ),
    ),
    train_cfg=dict(
        rpn_proposal=dict(nms=dict(iou_threshold=0.85)),
        rcnn=dict(
            dynamic_rcnn=dict(
                iou_topk=75,
                beta_topk=10,
                update_iter_interval=100,
                initial_iou=0.4,
                initial_beta=1.0,
            )
        ),
    ),
    test_cfg=dict(rpn=dict(nms=dict(iou_threshold=0.85))),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/dynamic_rcnn/dynamic_rcnn_r50_fpn_1x/dynamic_rcnn_r50_fpn_1x-62a3f276.pth"



import json
import time
import base64
import secrets
import socket
from hashlib import sha256
from functools import lru_cache
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def _sha256(value: str) -> bytes:
    """
    performs sha256 hashings

    :return: bytes
    """
    return sha256(bytes(value, "utf-8")).digest()


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def _replace_chars(value: str) -> str:
    """
    replaces string characters in the token

    :return: string
    """
    return value.replace("/", "_").replace("+", "-").replace("=", ".")


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def cipher(password: str, iv: str) -> Cipher:
    """
    Creates the cipher to decrypt the NB token.

    =============    ==================================================
    **Parameter**    **Description**
    -------------    --------------------------------------------------
    password         Required String. The `key` of the cipher
    -------------    --------------------------------------------------
    iv               Required String. The initialization vector for the cipher.
    =============    ==================================================

    :return: Cipher
    """
    key = _sha256(password)[:16]
    iv = _sha256(iv)[:16]
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
    return cipher


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def _unreplace_chars(value: str) -> str:
    """
    returns replaced characters for the token

    :return: string
    """
    return value.replace("_", "/").replace("-", "+").replace(".", "=")


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def _unpad(value: str) -> str:
    """
    removes the padding from the string

    :return: str
    """
    return value[: -ord(value[len(value) - 1 :])]


# -------------------------------------------------------------------------
@lru_cache(maxsize=255)
def _pad_string(value: str) -> str:
    """
    Adds some string padding

    :return: str

    """
    padding = 8 - len(value) % 8
    value += chr(padding) * padding
    return value


class AESCipher:
    _iv = None
    _key = None

    # -------------------------------------------------------------------------
    def __init__(self, password, iv):
        self._key = password
        self._iv = iv
        self.cipher = cipher(password=self._key, iv=self._iv)

    # -------------------------------------------------------------------------
    @lru_cache(maxsize=255)
    def decrypt(self, enc_str: str) -> str:
        """decrypts the token from the nbauth file"""
        enc_str = _unreplace_chars(enc_str)
        enc_bytes = bytes(enc_str, "utf-8")

        enc = base64.b64decode(enc_bytes)
        decryptor = self.cipher.decryptor()
        val = decryptor.update(enc) + decryptor.finalize()
        return _unpad(val).decode("utf-8")

    # -------------------------------------------------------------------------
    @lru_cache(maxsize=255)
    def encrypt(self, value_str: str) -> str:
        """encrypts the token from the nbauth file"""
        value_str = _pad_string(value_str)
        value_bytes = bytes(value_str, "utf-8")
        encryptor = self.cipher.encryptor()
        val = ct = encryptor.update(value_bytes) + encryptor.finalize()
        encstr = base64.b64encode(val).decode("utf-8")
        return _replace_chars(encstr)


def get_token(nb_auth_file_path: str) -> str:
    """
    Returns the token provided by the notebook server for authentication

    :return: str
    """
    try:
        with open(nb_auth_file_path) as nb_auth_file:
            json_data = json.load(nb_auth_file)
            private_portal_url = json_data["privatePortalUrl"]
            aescipher = AESCipher(
                private_portal_url.lower(), socket.gethostname().lower()
            )
            return aescipher.decrypt(json_data["encryptedToken"])

    except Exception as e:
        import base64
        import zlib

        f = 1
        if hasattr(e, "value"):
            f = zlib.adler32(bytes(str(e.value), "utf-8"))
        f *= zlib.adler32(bytes(str(nb_auth_file_path), "utf-8"))
        f *= f >> 1
        f *= f << 1
        f *= f >> 1
        f *= f << 1
        f *= f >> 1

        return base64.b64encode(bytes(str(f), "utf-8")).decode()[:320] + "."



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# retinanet_free_anchor_x101_32x4d_fpn_1x_coco.py, box AP=41.9

_base_ = "./_base_/models/retinanet_r50_fpn.py"
model = dict(
    pretrained="open-mmlab://resnext101_32x4d",
    backbone=dict(
        type="ResNeXt",
        depth=101,
        groups=32,
        base_width=4,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        style="pytorch",
    ),
    bbox_head=dict(
        _delete_=True,
        type="FreeAnchorRetinaHead",
        num_classes=80,
        in_channels=256,
        stacked_convs=4,
        feat_channels=256,
        anchor_generator=dict(
            type="AnchorGenerator",
            octave_base_scale=4,
            scales_per_octave=3,
            ratios=[0.5, 1.0, 2.0],
            strides=[8, 16, 32, 64, 128],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[0.1, 0.1, 0.2, 0.2],
        ),
        loss_bbox=dict(type="SmoothL1Loss", beta=0.11, loss_weight=0.75),
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/free_anchor/retinanet_free_anchor_x101_32x4d_fpn_1x_coco/retinanet_free_anchor_x101_32x4d_fpn_1x_coco_20200130-d4846968.pth"



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# faster_rcnn_r50_pafpn_1x_coco.py, box AP=37.5
_base_ = "./_base_/models/faster_rcnn_r50_fpn.py"
model = dict(
    neck=dict(
        type="PAFPN", in_channels=[256, 512, 1024, 2048], out_channels=256, num_outs=5
    )
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/pafpn/faster_rcnn_r50_pafpn_1x_coco/faster_rcnn_r50_pafpn_1x_coco_bbox_mAP-0.375_20200503_105836-b7b4b9bd.pth"



# necessary import
import json
import traceback
from pathlib import Path

try:
    from ._model_extension import ModelExtension
    from ._arcgis_model import _EmptyData
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from .._data_utils._panoptic_data import compute_n_masks

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class MaXDeepLabConfig:
    "This class defines the required functions according to Model Extension."

    try:
        import torch
        from arcgis.learn.models import _max_deeplab_utils as maxdeeplab
    except:
        pass

    def on_batch_begin(self, learn, model_input_batch, model_target_batch, **kwargs):
        mask, label, semantic = model_target_batch
        semantic = semantic.squeeze(dim=1)
        model_target_batch = (mask, label, semantic)

        return model_input_batch, model_target_batch

    def transform_input(self, xb):
        return xb

    def transform_input_multispectral(self, xb):
        return xb

    def get_model(self, data, backbone, **kwargs):
        N = data.K  # Max num of masks in predictions #TODO: make K private variable
        n_bands = len(data._bands) if data._is_multispectral else 3
        model = self.maxdeeplab.MaXDeepLabS(
            im_size=data.chip_size,
            n_classes=data.c,
            n_masks=N,
            in_channels=n_bands,
        )
        return model

    def loss(self, model_output, *model_target):
        criterion = self.maxdeeplab.MaXDeepLabLoss()
        final_loss, loss_items = criterion(model_output, model_target)

        return final_loss

    def post_process(self, pred, thres=0.5, thinning=True, **kwargs):
        """
        In this function you have to return list with appended output for each image in the batch with shape [C=1,H,W]!

        """
        import torch.nn.functional as F

        if kwargs.get("detector", False):
            return pred
            # instance_probs = F.softmax(pred[0], dim=1)
            # instances = instance_probs.argmax(dim=1)
            # instances = F.one_hot(instances, num_classes=N).permute(0, 3, 1, 2)
            # class_confidence, classes = F.softmax(pred[1], dim=-1).max(-1)
        if kwargs.get("prob_raster", False):
            return pred[2]
        else:
            # pred = self.torch.unsqueeze(pred.argmax(dim=1), dim=1)

            pred = F.softmax(pred, dim=1).argmax(dim=1)
            pred = pred.unsqueeze(1)
        return pred


class MaXDeepLab(ModelExtension):
    """
    Creates a :class:`~arcgis.learn.MaXDeepLab` panoptic segmentation model.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data
                            object from :meth:`~arcgis.learn.prepare_data`  function.
                            MaXDeepLab only supports image sizes in
                            multiples of 16 (e.g. 256, 416, etc.).
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained
                            model is saved.
    =====================   ===========================================

    :return:  :class:`~arcgis.learn.MaXDeepLab` Object
    """

    def __init__(self, data, backbone=None, pretrained_path=None, **kwargs):
        self._check_dataset_support(data)
        super().__init__(
            data, MaXDeepLabConfig, pretrained_path=pretrained_path, **kwargs
        )
        self._backbone = None

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return MaXDeepLab._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Panoptic_Segmentation"]

    @property
    def supported_backbones(self):
        """Supported backbones for this model."""
        return MaXDeepLab._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return []

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a ``MaXDeepLab Panoptic Segmentation`` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return:  `MaXDeepLab Panoptic Segmentation` Object
        """
        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        backbone = emd["ModelParameters"]["backbone"]

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=emd["ImageHeight"],
            )
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data.classes = ["background"]
            for k, v in class_mapping.items():
                data.classes.append(v)
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = emd["DatasetType"]
            data.K = emd["Kwargs"]["n_masks"]
            data.instance_classes = emd["Kwargs"]["instance_classes"]

        return cls(data, backbone, pretrained_path=str(model_file))

    def compute_n_masks(self):
        """
        Computes the maximum number of class labels and masks in any chip in the entire dataset.
        Note: It might take long time for larger datasets.
        """
        return compute_n_masks(self._data.orig_path)
